from sentenai.api import *
from sentenai.stream.events import Event

class Patterns(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, "patterns")

    def __getitem__(self, key):
        resp = self._get(key)
        if resp.status_code == 200:
            return Pattern.from_json(self, resp.json())
        elif resp.status_code == 404:
            raise KeyError("Pattern not found.")
        else:
            raise Exception(resp.status_code)

    def __delitem__(self, key):
        x = self[key]
        if x:
            resp = self._delete(key)
            if resp.status_code not in (200, 204):
                raise Exception(resp.status_code)
        else:
            raise KeyError("Pattern does not exist")

    def create(self, name, definition, description=""):
        if not definition:
            raise ValueError("Pattern must not be empty")
        else:
            resp = self._put(name, json={"pattern": definition if isinstance(definition, str) else {'select': definition.json()}, "description": description})
            if resp.status_code in [200, 201]:
                return self[name]
            else:
                raise Exception(resp.status_code)

    def __iter__(self):
        return iter([(x['name'], Pattern.from_json(self, x)) for x in self._get().json()])

    def __repr__(self):
        return API.__repr__(self) + ".patterns"

    def __call__(self, definition):
        if isinstance(definition, str):
            sj = definition
        elif isinstance(definition, bytes):
            sj = definition.decode('utf-8')
        else:
            sj = {'select': definition.json()}
        resp = self._post(json={"pattern": sj})
        if resp.status_code in [200, 201]:
            vid = resp.headers['Location']
            return Pattern(parent=self, name=vid, definition=sj, anonymous=True)
        else:
            raise Exception(resp.status_code)

class Pattern(API):
    def __init__(self, parent, name=None, description=None, definition=None, anonymous=False):
        API.__init__(self, parent._credentials, *parent._prefix, name)
        self._name = name
        self._description = description
        self._search = Search(self)
        self._definition = definition
        self._parent = parent
        self._anonymous = anonymous
        if anonymous and description is not None:
            raise ValueError("Anonymous patterns cannot have descriptions.")

    def __repr__(self):
        return repr(self._parent) + '[{}]'.format(repr(self._name))

    @property
    def description(self):
        return None if self._anonymous else self._description

    @property
    def name(self):
        return None if self._anonymous else self._name

    @property
    def search(self):
        return self._search

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, definition):
        if self._anonymous:
            raise TypeError("Cannot change definition of anonymous pattern")
        if isinstance(definition, str):
            s = definition
        else:
            s = {'select': definition.json()}
        resp = self._put(json={'pattern': s, 'description': self._description})
        if resp.status_code == 200:
            self._definition = s
            return self
        else:
            raise Exception(resp.status_code)

    @staticmethod
    def from_json(parent, x):
        return Pattern(parent, x['name'], x['description'], x['query'])


class Search(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, "search")
        self._parent = parent

    def __getitem__(self, s):
        params = {}
        if s.start is not None:
            params["start"] = iso8601(s.start)
        if s.stop is not None:
            params["end"] = iso8601(s.stop)
        if s.step is not None:
            params["limit"] = int(s.step)

        results = []
        while True:
            resp = self._get(params=params)
            if resp.status_code == 200:
                for x in resp.json():
                    results.append(
                            Event(
                                ts=dt64(x['start']),
                                duration=dt64(x['end']) - dt64(x['start']),
                                data={
                                    'name': self._parent.name,
                                    'pattern': self._parent._definition,
                                    'description': self._parent._description
                                }
                            )
                    )
                try:
                    params['start'] = resp.headers['cursor']
                except KeyError:
                    return results

            else:
                raise Exception(resp.status_code)




from sentenai.api import API, Credentials, PANDAS
from sentenai.stream import Streams, Event
from sentenai.view import Views
from sentenai.pattern import Patterns
from sentenai.pattern.expression import InCircle
if PANDAS: import pandas as pd

from threading import Thread
import time
from queue import Queue, Empty

__all__ = ['Client', 'Event']

if PANDAS:
    def df(events):
        return pd.DataFrame([x.as_record() for x in events])



class Client(API):
    def __init__(self, host, auth_key=""):
        API.__init__(self, Credentials(host, auth_key))

    def __repr__(self):
        return "Client(auth_key=\"{}\", host=\"{}\")".format(self._credentials.auth_key, self._credentials.host)

    def ping(self):
        """float: Ping Sentenai cluster's health check and get back response time in seconds."""
        t0 = time.time()
        r = self._get()
        if r.status_code == 200:
            return time.time() - t0
        else:
            raise Exception(r.status_code)

    @property
    def streams(self):
        """Streams: the sub-api for streams."""
        return Streams(self)

    @property
    def views(self):
        """Views: the sub-api for views."""
        return Views(self)

    @property
    def patterns(self):
        """Patterns: the sub-api for patterns."""
        return Patterns(self)










from sentenai.stream.metadata import Metadata
from sentenai.stream.events import Events, Event
from sentenai.stream.fields import Fields, Field
from sentenai.api import API, iso8601, SentenaiEncoder, PANDAS
if PANDAS:
    import pandas as pd
from datetime import datetime
import simplejson as JSON
import re, io

# Optional for parquet handling
try:
    import pyarrow
except:
    pass


class Streams(API):
    def __init__(self, parent):
        self._parent = parent
        API.__init__(self, parent._credentials, *parent._prefix, "streams")


    if PANDAS:
        def _repr_html_(self):
            data = list(self)
            return pd.DataFrame([{'name': k, 't0': s.t0} for k, s in data])._repr_html_()

    def __repr__(self):
        return "{}.streams".format(repr(self._parent))

    def __getitem__(self, key):
        if type(key) is tuple:
            k, a = key
            return Stream(self, name=k, anchor=a)
        else:
            return Stream(self, name=key)

    def __len__(self):
        return len(self._get().json())

    def __delitem__(self, key):
        x = self[key]
        if x:
            x.delete()
        else:
            raise KeyError("Stream does not exist")

    def __iter__(self):
        return iter([(x['name'], Stream(self, name=x['name'], anchor=x.get('t0'))) for x in self._get().json()])
    
    def __call__(self, name=".*", **kwargs):
        ss = []
        for item in self._get().json():
            if not re.search(name, item['name']):
                continue
            for k, v in kwargs.items():
                if k not in item['meta'] or not re.search(v, item['meta'][k]):
                    break
            else:
                ss.append(item)
        return iter([(x['name'], Stream(self, name=x['name'])) for x in ss])
        
        




class Stream(API):
    def __init__(self, parent, name, filters=None, anchor=None):
        p = {'filters': filters.json()} if filters else {}
        API.__init__(self, parent._credentials, *parent._prefix, name, params=p)
        self._parent = parent
        self._name = name
        self._filters = filters
        self._anchor = anchor

    def init(self, t0="now"):
        if t0 == "now":
            r = self._put(headers={'t0': iso8601(datetime.utcnow())}, json=None)
        elif t0 == None:
            r = self._put()
        else:
            r = self._put(headers={'t0': iso8601(t0)}, json=None)
        if r.status_code != 201:
            raise Exception(r.status_code)

    def upload(self, events):
        r = None
        hdr = {'content-type': 'application/x-ndjson'}
        if isinstance(events, str):
            with open(events) as f:
                r = self._post(json=f.read(), headers=hdr)
        elif isinstance(events, io.IOBase):
            r = self._post(json=events.read(), headers=hdr)
        elif isinstance(events, pyarrow.Table):
            cnames = list(map(str.lower, events.column_names))
            if 'ts' in cnames:
                ts_ix = cnames.index('ts')
            elif 'timestamp' in cnames:
                ts_ix = cnames.index('timestamp')
            else:
                raise ValueError("Needs timestamp column")
            if 'duration' in cnames:
                td_ix = cnames.index('ts')
            else:
                td_ix = None

            def iterate():
                for n in range(len(events)):
                    e = Event(ts=events[ts_ix][n].as_py(), duration=events[td_ix][n].as_py() if td_ix else None, data={})
                    for i, c in enumerate(events.column_names):
                        if i in [td_ix, ts_ix]:
                            continue
                        e.data[events.column_names[i]] = events[i][n].as_py()
                    yield (
                        JSON.dumps({
                            "id": e.id,
                            "ts": e.ts,
                            "duration": e.duration,
                            "event": e.data
                            }, ignore_nan=True, cls=SentenaiEncoder
                        ) + '\n'
                    ).encode()
            r = self._post(json=iterate())
        else:
            r = self._post(json=((JSON.dumps({"id": e.id, "ts": e.ts, "duration": e.duration, "event": e.data}, ignore_nan=True, cls=SentenaiEncoder)+"\n").encode() for e in events), headers=hdr)
        if r.status_code not in range(200, 300):
            raise Exception(r.status_code)

    def __repr__(self):
        return 'Stream(name={!r})'.format(self._name)

    def __delattr__(self, name):
        if name == 'metadata':
            self.metadata.clear()
        else:
            raise TypeError("cannot delete `{}`".format(name))

    def where(self, *filters):
        if len(filters) < 1:
            fs = self._filters
        elif self._filters:
            fs = self._filters & filters[0]
        else:
            fs = filters[0]
        for f in filters[1:]:
            fs &= f
        return Stream(self._parent, self._name, fs, self._anchor)
            
    def json(self):
        d = {'name': self._name}
        if self._anchor:
            d['t0'] = self._anchor
        if 'filters' in self._params:
            d['filter'] = self._params['filters']
        return d

    @property
    def metadata(self):
        return Metadata(self)

    @property
    def events(self):
        return Events(self)

    @property
    def bounds(self):
        try:
            return self.events[::1][0].ts, self.events[::-1][0].ts
        except:
            return (None, None)

    @property
    def t0(self):
        if self._anchor:
            return self._anchor
        else:
            return self._head().headers.get('t0')

    @property
    def name(self):
        return self._name

    def update(self, evt):
        return self.events.update(evt)

    def insert(self, evt):
        return self.events.insert(evt)

    def remove(self, evt):
        return self.events.delete(evt)

    def values(self, at=None):
        params = {}
        if self._anchor:
            params['t0'] = iso8601(self._anchor)
        if at:
            if self.t0:
                params['at'] = iso8601(at)
            else:
                params['at'] = at

        res = self._get(params=params)
        if res.status_code == 200:
            vs = res.json()
            return dict([(tuple(x['path']) if len(x['path']) > 1 else x['path'][0], x['value']) for x in vs])
        elif res.status_code == 404:
            raise ValueError("stream not found")
        else:
            raise Exception(res.status_code)

    def __matmul__(self, ts):
        return self.values(ts)

    def delete(self):
        r = self._delete()
        if r.status_code == 204:
            return None
        else:
            raise Exception("Couldn't delete")

    def __bool__(self):
        r = self._head()
        if r.status_code == 404:
            return False
        elif r.status_code == 200:
            return True
        else:
            raise Exception(r.status_code)

    def __nonzero__(self):
        r = self._head()
        if r.status_code == 404:
            return False
        elif r.status_code == 200:
            return True
        else:
            raise NotImplemented

    def __str__(self):
        return f'stream "{self._name!s}"'

    @property
    def fields(self):
        return Fields(self)

    def __iter__(self):
        return iter(Fields(self))

    def __getitem__(self, key):
        return Fields(self)[(key,) if type(key) != tuple else key]

    def __setitem__(self, key, val):
        raise NotImplemented("What would it do?")







from sentenai.api import API, iso8601, dt64, td64, PANDAS
from sentenai.view.expression import Var
from sentenai.view.views import Views
from sentenai.pattern.expression import Path
if PANDAS: import pandas as pd

class Fields(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, params=parent._params)
        self._stream = parent

    def __repr__(self):
        return repr(self._stream) + ".fields"

    if PANDAS:
        def _repr_html_(self):
            return pd.DataFrame([
                {'field': str(f)} for f in iter(self)
            ])._repr_html_()

    def __iter__(self):
        x = self._get("fields")
        if x.status_code == 200:
            return iter(Field(self, self._stream, *f['path']) for f in x.json())
        elif x.status_code == 404:
            raise ValueError("stream not found.")
        else:
            raise Exception(x.status_code)

    def __getitem__(self, key):
        if type(key) == tuple:
            return Field(self, self._stream, *key)
        elif type(key) == slice:
            if key.start is None and key.stop == 'ts':
                return TimestampField(self, self._stream)
            elif key.start is None and key.stop == 'duration':
                return DurationField(self, self._stream)
            else:
                raise ValueError("invalid path")
        else:
            return Field(self, self._stream, key)

class FieldData(object):
    def __init__(self, stream, do):
        self._str = stream
        self._do = do

    def __getitem__(self, s):
        if isinstance(s, slice) and self._str.t0 and (s.start is None or s.end is None):
            t0, t1 = self._str.bounds
            s = slice(s.start or t0, s.stop or t1, s.step)
        return self._do[s]







class Field(API, Var, Path):
    def __init__(self, parent, stream, *path, **kwargs):
        API.__init__(self, parent._credentials, *parent._prefix, params=parent._params)
        self._parent = parent
        self._path = path
        self._shift = None
        self._stream = stream
        self._start = kwargs.get('start')
        self._end = kwargs.get('end')

    def __repr__(self):
        return "{}[{}]".format(repr(self._stream), ', '.join(['"{}"'.format(x) for x in self._path]))

    def __str__(self):
        return "/".join(self._path)

    @property
    def data(self):
        d = Views(API(self._credentials))(value=self).data
        return FieldData(self._stream, d)


    def json(self):
        d = {'path': ("event", ) + self._path, 'stream': self._stream.json()}
        if self._shift:
            d['shift'] = self._shift
        return d

    @property
    def path(self):
        return tuple(self._path)

    def __matmul__(self, ts):
        return self.value(at=ts)

    def value(self, at=None):
        res = self._get("values", *self._path, params={'at': iso8601(at)} if at is not None else {})
        if res.status_code == 200:
            return res.json()['value']
        elif res.status_code == 404:
            raise ValueError("stream not found")
        else:
            raise Exception(res.status_code)

    def __getitem__(self, s):
        if type(s) == slice:
            return Field(self._parent, self._stream, *self._path, start=s.start, end=s.stop)

            params = {}
            if s.start is not None:
                try:
                    params['start'] = iso8601(dt64(s.start))
                except:
                    raise ValueError("Range slicing only allowed with datetime types")

            if s.stop is not None:
                try:
                    params['end'] = iso8601(dt64(s.stop))
                except:
                    raise ValueError("Range slicing only allowed with datetime types")



        else:
            raise ValueError("Field type can only use range slicing.")

    @property
    def stats(self): 
        params = {}
        if self._start:
            params['start'] = self._start
        if self._end:
            params['end'] = self._end
        if self._stream._filters:
            params['filters'] = self._stream._filters.json()

        resp = self._get('stats', *self._path, params=params)
        if resp.status_code == 200:
            data = resp.json()
            return FieldStats(data)
        elif resp.status_code == 404:
            raise ValueError("stream not found")
        else:
            raise Exception(resp.status_code)

    # shortcuts go get full range field stats
    @property
    def min(self): return self.stats.min
    @property
    def max(self): return self.stats.max
    @property
    def mean(self): return self.stats.mean
    @property
    def std(self): return self.stats.std
    @property
    def top(self): return self.stats.top
    @property
    def freq(self): return self.stats.freq
    @property
    def count(self): return self.stats.count
    @property
    def missing(self): return self.stats.missing

    @property
    def unique(self):
        params = {}
        if self._start:
            params['start'] = self._start
        if self._end:
            params['end'] = self._end
        params['filters'] = self._stream._filters.json() if self._stream._filters else None
        resp = self._get("uniques", *self._path, params=params)
        if resp.status_code == 200:
            u = resp.json()
            if not u['categorical']:
                return {float(v) for v in dict(u['numerical']).keys()}
            else:
                return set(dict(u['categorical']).keys())
        else:
            raise Exception(resp.status_code)

class TimestampField(Field):
    def json(self):
        d = {'path': ("ts", ) + self._path, 'stream': self._stream.json()}
        if self._shift:
            d['shift'] = self._shift
        return d

class DurationField(Field):
    def json(self):
        d = {'path': ("duration", ) + self._path, 'stream': self._stream.json()}
        if self._shift:
            d['shift'] = self._shift
        return d



class FieldStats(object):
    def __init__(self, stats):
        self._stats = stats

    if PANDAS:
        def _repr_html_(self):
            return pd.DataFrame([
                {'Statistic': k, 'Value': v}
                for k, v in self._stats[
                    'numerical'
                    if self._stats.get('numerical')
                    else 'categorical'
                ].items()
            ])._repr_html_()


    @property
    def mean(self):
        try:
            return self._stats['numerical']['mean']
        except KeyError:
            return None

    @property
    def min(self):
        try:
            return self._stats['numerical']['min']
        except KeyError:
            return None

    @property
    def max(self):
        try:
            return self._stats['numerical']['max']
        except KeyError:
            return None

    @property
    def std(self):
        try:
            return self._stats['numerical']['std']
        except:
            return None

    @property
    def missing(self):
        try:
            if self._stats['numerical']:
                return self._stats['numerical']['missing']
            elif self._stats['categorical']:
                return self._stats['categorical']['missing']
            else:
                return None
        except KeyError:
            return None

    @property
    def count(self):
        try:
            if self._stats['numerical']:
                return self._stats['numerical']['count']
            elif self._stats['categorical']:
                return self._stats['categorical']['count']
            else:
                return None
        except KeyError:
            return None

    @property
    def top(self):
        try:
            return self._stats['categorical']['top']
        except KeyError:
            return None

    @property
    def freq(self):
        try:
            return self._stats['categorical']['freq']
        except KeyError:
            return None



import json


class Expression(object):

    def __init__(self, fdict):
        self.fdict = {}
        for k, v in fdict.items():
            if isinstance(v, Formula):
                self.fdict[k] = v
            elif isliteral(v):
                self.fdict[k] = Literal(v)
            else:
                self.fdict[k] = Formula(v)

    def json(self):
        f = {}
        s = []
        for i, (k, v) in enumerate(self.fdict.items()):
            if isinstance(v, (Literal, Var)):
                f[k] = [v.json()]
            else:
                fn = 'feature-{:04d}'.format(i)
                f[k] = [{'var': ['event', fn]}]
                n = v.json(fn)
                s.append(n)

        #d = {'explicit': [{'type':'projection' if len(s) < 2 else 'outerjoin',  'fields':f,  'source':s}]}
        if len(s) > 1:
            return {'type': 'projection', 'fields': f, 'source': [{'type': 'outerjoin', 'source': s}]}
        else:
            return {'type':'projection',  'fields':f,  'source':s}


class Var(object):
    def __add__(self, other):
        return Formula(self) + other

    def __radd__(self, other):
        return other + Formula(self)

    def __sub__(self, other):
        return Formula(self) - other

    def __rsub__(self, other):
        return other - Formula(self)

    def __mul__(self, other):
        return Formula(self) * other

    def __rmul__(self, other):
        return other * Formula(self)

    def __truediv__(self, other):
        return Formula(self) / other

    def __rtruediv__(self, other):
        return other / Formula(self)

    def resample(self, freq):
        return Formula(self).resample(freq)

    def rolling(self, n, win_type=None):
        return Formula(self).rolling(n, win_type)

    def shift(self, n):
        return Formula(self).shift(n)

class Formula(object):

    def __init__(self, src):
        self._src = src
        self._shift = None

    def json(self, name='_'):
        j = self._src.json()
        d = {'var': j['path']}
        if self._shift:
            d['shift'] = self._shift
        return {'type':'projection', 
         'fields':{name: [d]}, 
         'source':[
          {'stream':j['stream'], 
           'projection':'default'}]}

    def shift(self, n):
        self._shift = n
        return self

    def resample(self, freq):
        return Resample(self, freq)

    def rolling(self, n, win_type=None):
        return Rolling(self, n, win_type)

    def __add__(self, n):
        if isinstance(n, Formula):
            return BinOp(self, '+', n)
        elif isliteral(n):
            return BinOp(self, '+', Literal(n))
        else:
            return BinOp(self, '+', Formula(n))

    def __mul__(self, n):
        if isinstance(n, Formula):
            return BinOp(self, '*', n)
        elif isliteral(n):
            return BinOp(self, '*', Literal(n))
        else:
            return BinOp(self, '*', Formula(n))

    def __rmul__(self, n):
        if isinstance(n, Formula):
            return BinOp(n, '*', self)
        elif isliteral(n):
            return BinOp(Literal(n), '*', self)
        else:
            return BinOp(Formula(n), '*', self)

    def __truediv__(self, n):
        if isinstance(n, Formula):
            return BinOp(self, '/', n)
        elif isliteral(n):
            return BinOp(self, '/', Literal(n))
        else:
            return BinOp(self, '/', Formula(n))

    def __rtruediv__(self, n):
        if isinstance(n, Formula):
            return BinOp(n, '/', self)
        elif isliteral(n):
            return BinOp(Literal(n), '/', self)
        else:
            return BinOp(Formula(n), '/', self)

    def __radd__(self, n):
        if isinstance(n, Formula):
            return BinOp(n, '+', self)
        elif isliteral(n):
            return BinOp(Literal(n), '+', self)
        else:
            return BinOp(Formula(n), '+', self)

    def __sub__(self, n):
        if isinstance(n, Formula):
            return BinOp(self, '-', n)
        elif isliteral(n):
            return BinOp(self, '-', Literal(n))
        else:
            return BinOp(self, '-', Formula(n))

    def __rsub__(self, n):
        if isinstance(n, Formula):
            return BinOp(n, '-', self)
        elif isliteral(n):
            return BinOp(Literal(n), '-', self)
        else:
            return BinOp(Formula(n), '-', self)


class Shift(Formula):

    def __init__(self, ft, n):
        self.n = n
        self.source = ft

    def json(self, name='_'):
        x = self.source.json(name)
        x['shift'] = self.n
        return x


class Resample(object):

    def __init__(self, ft, freq):
        self.source = ft
        self.freq = freq

    def max(self):
        return Resampled(self.source, self.freq, 'max')

    def min(self):
        return Resampled(self.source, self.freq, 'min')

    def mean(self):
        return Resampled(self.source, self.freq, 'mean')

    def std(self):
        return Resampled(self.source, self.freq, 'std')

    def sum(self):
        return Resampled(self.source, self.freq, 'sum')


class Resampled(Formula):

    def __init__(self, ft, freq, op):
        self.source = ft
        self.freq = freq
        self.op = op
        self.fill = 'na'
        self._offset = 0
        self._shift = None

    def offset(td):
        self._offset = td.total_seconds()
        return self

    def ffill(self):
        self.fill = 'ffill'
        return self

    def json(self, name='_'):
        j = {}
        v = {'var': ['event', '_']}
        if self._shift:
            v['shift'] = self._shift
        j['frequency'] = self.freq
        j['offset'] = self._offset
        j['type'] = 'resample'
        j['fill'] = self.fill
        j['source'] = [self.source.json()]
        j['fields'] = {name: {'expr':[v],  'aggregation':self.op}}
        return j


class Rolling(object):

    def __init__(self, ft, n, win_type='none', center=False):
        self.source = ft
        self.n = n
        self.center = center
        self.win_type = win_type

    def max(self, **kwargs):
        kwargs['type'] = self.win_type
        if self.win_type == "gaussian" and 'std' not in kwargs:
            kwargs['std'] = 1
        return Windowed(self, 'max', kwargs)

    def min(self, **kwargs):
        kwargs['type'] = self.win_type
        if self.win_type == "gaussian" and 'std' not in kwargs:
            kwargs['std'] = 1
        return Windowed(self, 'min', kwargs)

    def mean(self, **kwargs):
        kwargs['type'] = self.win_type
        if self.win_type == "gaussian" and 'std' not in kwargs:
            kwargs['std'] = 1
        return Windowed(self, 'mean', kwargs)

    def std(self, **kwargs):
        kwargs['type'] = self.win_type
        if self.win_type == "gaussian" and 'std' not in kwargs:
            kwargs['std'] = 1
        return Windowed(self, 'std', kwargs)

    def sum(self, **kwargs):
        kwargs['type'] = self.win_type
        if self.win_type == "gaussian" and 'std' not in kwargs:
            kwargs['std'] = 1
        return Windowed(self, 'sum', kwargs)


class Windowed(Formula):

    def __init__(self, rolling, op, winargs):
        self.win_args = winargs
        self.op = op
        self.r = rolling
        self._shift = None

    def json(self, name='_'):
        x = {'window':self.win_args, 
             'size':self.r.n, 
             'op':self.op, 
             'var':['event', '_'], 
             'labelPos':'center' if self.r.center else 'right'
             }
        if self._shift:
            x['shift'] = self._shift
        return {'fields':{name: [x]}, 
         'source':[
          self.r.source.json()], 
         'type':'projection'}


def isliteral(v):
    return type(v) in (float, int, str, bool)


class Literal(Formula):

    def __init__(self, val):
        self.val = val

    def shift(self):
        raise TypeError("Can't shift a literal.")

    def json(self, name='_'):
        x = self.val
        if isinstance(x, float):
            return {'lit': {'val':x,  'type':'double'}}
        if isinstance(x, int):
            return {'lit': {'val':x,  'type':'int'}}
        if isinstance(x, bool):
            return {'lit': {'val':x,  'type':'int'}}
        if isinstance(x, str):
            return {'lit': {'val':x,  'type':'str'}}
        raise TypeError('unsupported')


class BinOp(Formula):

    def __init__(self, left, op, right):
        self._left = left
        self._right = right
        self._op = op

    def json(self, name='_'):
        src = []
        if isinstance(self._left, Literal):
            lhs = self._left.json()
        else:
            lhs = {'var': ['event', '_lhs']}
            src.append(self._left.json('_lhs'))
        if isinstance(self._right, Literal):
            rhs = self._right.json()
        else:
            rhs = {'var': ['event', '_rhs']}
            src.append(self._right.json('_rhs'))
        if len(src) > 1:
            src = [{'source': src, 'type': 'outerjoin'}]
        return {'fields':{name: [{'lhs':lhs,  'rhs':rhs,  'op':self._op}]},  'source':src,  'type': 'projection'}



from datetime import date, datetime
class Expression(object):

    @property
    def followed_by(self):
        return Function(Seq, self)

    @property
    def before(self):
        return Function(Seq, self)

    @property
    def during(self):
        return Function(During, self)

    @property
    def union(self):
        return Function(Or, self)

    @property
    def intersection(self):
        return Function(And, self)

    def __or__(self, other):
        return Or(self, other)

    def __and__(self, other):
        return And(self, other)

    def __truediv__(self, other):
        return During(self, other)

    def __rshift__(self, other):
        return Seq(self, other)


class Path(object):
    def __eq__(self, other):
        if isinstance(other, InCircle):
            return Condition(self, 'in', other)
        else:
            return Condition(self, '==', other)
    def __lt__(self, other):
        return Condition(self, '<', other)
    def __gt__(self, other):
        return Condition(self, '>', other)
    def __ne__(self, other):
        return Condition(self, '!=', other)
    def __le__(self, other):
        return Condition(self, '<=', other)
    def __ge__(self, other):
        return Condition(self, '>=', other)


class When(Expression):
    def __init__(self, signal, duration, within):
        self._signal = signal
        self._within = within
        self._duration = duration

    def json(self):
        if isinstance(self._signal, When):
            d = {'type':'serial', 
             'conds':[self._signal.json()]}
        else:
            d = self._signal.json()
        if self._duration:
            d.update(self._duration.json())
        if self._within:
            d.update(self._within.json())
        return d

class InCircle(object):
    def __init__(self, pt, radius):
        self._pt = pt
        self._r = radius

    def json(self):
        return {
            'center': { 'lon': self._pt[0], 'lat': self._pt[1] },
            'radius': self._r
        }

class Condition(When):

    def __init__(self, var, op, val):
        self._var = var
        self._op = op
        self._val = val

    def json(self):
        val = self._val
        if isinstance(self._val, float):
            vt = 'double'
        elif isinstance(self._val, bool):
            vt = 'bool'
        elif isinstance(self._val, int):
            vt = 'double'
        elif isinstance(self._val, date):
            vt = 'date'
        elif isinstance(self._val, datetime):
            vt = 'datetime'
        elif isinstance(self._val, InCircle):
            vt = 'circle'
            val = self._val.json()
        else:
            vt = 'string'
        d = {'op':self._op, 'arg':{'type':vt,  'val':val},  'type':'span'}
        d.update(self._var.json())
        return d


class Or(When):

    def __init__(self, *args):
        if len(args) == 0:
            raise ValueError("At least one condition required")
        self._args = list(args)

    def json(self):
        if len(self._args) == 1:
            return self._args[0].json()

        top = {'expr': '||', 'args': [self._args[0].json()]}
        last = top
        for x in self._args[1:-1]:
            z = {'expr': '||', 'args': [x.json()]}
            last['args'].append(z)
            last = z
        last['args'].append(self._args[-1].json())
        return top

class And(When):

    def __init__(self, *args):
        self._args = list(args)

    def json(self):
        top = {'expr': '&&', 'args': [self._args[0].json()]}
        last = top
        for x in self._args[1:-1]:
            z = {'expr': '&&', 'args': [x.json()]}
            last['args'].append(z)
            last = z
        last['args'].append(self._args[-1].json())
        return top


class During(When):

    def __init__(self, x, y):
        self._contained = x
        self._container = y

    def json(self):
        return {'type':'during', 
         'conds':[self._contained.json(), self._container.json()]}


class Either(When):

    def __init__(self, signal1, signal2):
        self._signals = [
         signal1, signal2]

    def json(self):
        return {'type':'any', 
         'conds':[s.json() for s in self._signals]}


class Any(When):

    def __init__(self, *signals):
        self._signals = signals

    def json(self):
        return {'type':'any', 
         'conds':[s.json() for s in self._signals]}


class All(When):

    def __init__(self, *signals):
        self._signals = signals

    def json(self):
        return {'type':'all', 
         'conds':[s.json() for s in self._signals]}


class Seq(When):

    def __init__(self, *signals):
        self._signals = list(signals)

    def json(self):
        if len(self._signals) == 0:
            raise ValueError("Can't have zero-length sequence")
        elif len(self._signals) == 1:
            return self._signals[0].json()
        else:
            return {'type':'serial', 
             'conds':[s.json() for s in self._signals]}

    def __getitem__(self, s):
        if isinstance(s, tuple):
            return Seq(*self._signals, *s)
        if isinstance(s, When):
            return Seq(*self._signals, *(s,))
        if isinstance(s, slice):
            return Seq(*self._signals, *(When(s.start, s.stop, s.step),))
        raise TypeError("Can't get that type")


class Function(object):

    def __init__(self, ftype, left=None):
        self._left = left
        self._type = ftype

    def __getitem__(self, slices):
        if isinstance(slices, When):
            s = [
             slices]
        elif isinstance(slices, slice):
            s = [
             When(slices.start, slices.stop, slices.step)]
        elif isinstance(slices, tuple):
            s = []
            for sl in slices:
                if type(sl) is slice:
                    s.append(When(sl.start, sl.stop, sl.step))
                elif isinstance(sl, When):
                    s.append(sl)
                else:
                    raise TypeError('Invalid type in query')

        else:
            raise TypeError('Invalid type in query')
        if len(s) == 0:
            raise ValueError('Cannot be empty')
        if self._left is None:
            if len(s) == 1:
                return s[0]
            if self._type is Seq:
                return Seq(*s)
            raise ValueError("left can't be empty")
        elif isinstance(self._type, Seq) or len(s) == 1:
            return (self._type)(self._left, *s)
        else:
            return self._type(self._left, Seq(*s))


class Within(object):

    def __init__(self, **kwargs):
        self._args = kwargs
        self._dir = ['exact', 'within', 'after']
        self._pos = 1

    def json(self):
        return {self._dir[self._pos]: self._args}

    def __invert__(self):
        self._pos *= -1
        return self


class Duration(object):

    def __init__(self, **kwargs):
        self._args = kwargs


class Max(Duration):

    def json(self):
        return {'for': {'at-most': self._args}}

    def __and__(self, other):
        if not isinstance(other, Min):
            raise ValueError('Can only combine Max with Min')
        else:
            return MinMax(self, other)


class Min(Duration):

    def json(self):
        return {'for': {'at-least': self._args}}

    def __and__(self, other):
        if not isinstance(other, Max):
            raise ValueError('Can only combine Min with Max')
        else:
            return MinMax(self, other)


class MinMax(Duration):

    def __init__(self, min, max):
        self._min = min
        self._max = max

    def json(self):
        return {'for': {'at-least':self._min.json()['for']['at-least'],  'at-most':self._max.json()['for']['at-most']}}
# okay decompiling __pycache__/signal.cpython-37.pyc



from shapely.geometry import Point
from datetime import datetime
from sentenai.api import API, dt64, PANDAS
if PANDAS: import pandas as pd

class Metadata(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, 'metadata')
        self._parent = parent

    def clear(self):
        resp = self._put(json={})
        if resp.status_code not in (200, 204):
            raise Exception(resp.status_code)

    def __iter__(self):
        resp = self._get()
        if resp.status_code == 404:
            return None
        elif resp.status_code == 200:
            data = resp.json()
            parsed = {}
            for k,v in data.items():
                if type(v) in [float, int, bool]:
                    parsed[k] = v
                elif type(v) == dict and 'lat' in v and 'lon' in v:
                    parsed[k] = Point(v['lon'], v['lat'])
                else:
                    for fmt in ["%Y-%m-%dT%H:%M:%S.%fZ","%Y-%m-%dT%H:%M:%SZ","%Y-%m-%dT%H:%M:%S","%Y-%m-%dT%H:%M:%S.%f"]:
                        try:
                            val = dt64(datetime.strptime(v, fmt))
                        except ValueError:
                            pass
                        else:
                            parsed[k] = val
                            break
                    else:
                        parsed[k] = v

            return iter(parsed.items())
        else:
            raise Exception(resp.status_code)


    def __repr__(self):
        return repr(self._parent) + ".metadata"

    if PANDAS:
        def _repr_html_(self):
            return pd.DataFrame([
                {'key': n, 'value': v} for n, v in iter(self)
            ])._repr_html_()



    def _type(self, v):
        if type(v) in [int, float]:
            return "Numeric"
        elif type(v) == datetime:
            return "Datetime"
        elif type(v) == bool:
            return "Boolean"
        else:
            return "String"

    def __getitem__(self, key):
        return dict(self)[key]

    def __setitem__(self, key, val):
        resp = self._patch(json={key: val})
        if resp.status_code not in [200, 201, 204]:
            raise Exception(resp.status_code)

    def __delitem__(self, key):
        resp = self._delete(key)
        if resp.status_code not in [200, 204]:
            raise Exception(resp.status_code)



from sentenai.api import *
from sentenai.view.expression import Expression
from numpy import datetime64, timedelta64
from sentenai.stream.events import Event


class Views(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, "views")

    def __getitem__(self, key):
        resp = self._get(key)
        if resp.status_code == 200:
            return View.from_json(self, resp.json())
        elif resp.status_code == 404:
            raise KeyError("View not found.")
        else:
            raise Exception(resp.status_code)

    def __delitem__(self, key):
        x = self[key]
        if x:
            resp = self._delete(key)
            if resp.status_code not in (200, 204):
                raise Exception(resp.status_code)
        else:
            raise KeyError("View does not exist")

    def create(self, name, definition, description=""):
        if not definition:
            raise ValueError("View must not be empty")
        else:
            if type(definition) == str:
                vd = definition
            else:
                vd = Expression(definition).json()
            resp = self._put(name, json={"view": vd, "description": description})
            if resp.status_code in [200, 201]:
                return self[name]
            else:
                raise Exception(resp.status_code)

    def __iter__(self):
        return iter([(x['name'], View.from_json(self, x)) for x in self._get().json()])

    def __repr__(self):
        return API.__repr__(self) + ".views"

    def __call__(self, *fields, **definition):
        if not definition and len(fields) == 1 and type(fields[0]) == str:
            # text view
            resp = self._post(json={"view": fields[0]})
            if resp.status_code in [200, 201]:
                vid = resp.headers['Location']
                return View(parent=self, name=vid, definition=fields[0], anonymous=True)
            else:
                raise Exception(resp.status_code)

        from sentenai.stream.fields import Field
        for field in fields:
            if not isinstance(field, Field):
                raise TypeError("Default named fields must be instance of `Field` class.")
            d = definition
            for ps in field.path:
                if ps in d and isinstance(d[ps], dict):
                    d = d[ps]
                elif ps not in d:
                    d[ps] = {}
                else:
                    raise ValueError("Field named `{}` defined in both `fields` and `definition`.")
            else:
                d[ps] = field
        if not definition:
            raise ValueError("View must not be empty")
        else:
            sj = Expression(definition).json()
            resp = self._post(json={"view": sj})
            if resp.status_code in [200, 201]:
                vid = resp.headers['Location']
                return View(parent=self, name=vid, definition=sj, anonymous=True)
            else:
                raise Exception(resp.status_code)



class View(API):
    def __init__(self, parent, name=None, description=None, definition=None, anonymous=False):
        API.__init__(self, parent._credentials, *parent._prefix, name)
        self._name = name
        self._description = description
        self._data = Data(self)
        self._definition = definition
        self._parent = parent
        self._anonymous = anonymous
        if anonymous and description is not None:
            raise ValueError("Anonymous views cannot have descriptions.")

    def __repr__(self):
        return repr(self._parent) + '[{}]'.format(repr(self._name))

    @property
    def description(self):
        return None if self._anonymous else self._description

    @property
    def name(self):
        return None if self._anonymous else self._name

    @property
    def data(self):
        return self._data

    def __matmul__(self, at):
        at = dt64(at)
        z = self._data[at:at - datetime(1970,1,1):1]
        if len(z) == 0:
            return None
        else:
            z[0].ts = at
            z[0].duration = None
            return z[0]

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, definition):
        if self._anonymous:
            raise TypeError("Cannot change definition of anonymous view")
        s = Expression(definition).json()
        resp = self._put(json={'view': s, 'description': self._description})
        if resp.status_code == 200:
            self._definition = s
            return self
        else:
            raise Exception(resp.status_code)

    @staticmethod
    def from_json(parent, x):
        return View(parent, x.get('name', ''), x.get('description'))

class Data(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, "data")

    def __getitem__(self, s):

        # handle tuple of time ranges
        if isinstance(s, (list, tuple)):
            return [self[x] for x in s]

        params = {'sort': 'asc'}
        stype = None
        if isinstance(s, Event):
            e = s
            if e.duration is None:
                raise ValueError("Cannot calculate end time.")
            else:
                params['start'] = e.ts
                params['end'] = e.ts + e.duration
        elif isinstance(s, slice) and isinstance(s.start, Event):
            e = s.start
            if e.duration is None:
                raise ValueError("Cannot calculate end time.")
            else:
                params['start'] = e.ts
                params['end'] = e.ts + e.duration
            if s.stop is not None:
                params['limit'] = s.stop
            if s.step is not None:
                raise TypeError("3 segment slice illegal with event start.")
        else:
            if s.start is None or s.stop is None:
                raise TypeError("View data *must* include valid time window")

            # handle anchored time deltas
            if isinstance(s.start, (timedelta, timedelta64)):
                if isinstance(s.stop, (datetime, datetime64)):
                    params['end'] = s.stop
                    params['start'] = s.stop + s.start
                else:
                    raise TypeError("slice.stop must be datetime type when slice.start is None or timedelta")
            elif isinstance(s.stop, (timedelta, timedelta64)):
                if isinstance(s.start, (datetime, datetime64)):
                    params['start'] = s.start
                    params['end'] = s.start + s.stop
                else:
                    raise TypeError("slice.start must be datetime type when slice.stop is None or timedelta")
            else:
                if s.start is not None:
                    if not isinstance(s.start, (datetime, datetime64)):
                        raise TypeError("slice start must be either None or datetime/timedelta.")
                    params['start'] = s.start
                if s.stop is not None:
                    if not isinstance(s.stop, (datetime, datetime64)):
                        raise TypeError("slice stop must be either None or datetime/timedelta.")
                    params['end'] = s.stop


            # reverse time order if necessary
            if 'start' in params and 'end' in params:
                if params['start'] > params['end']:
                    params['start'], params['end'] = params['end'], params['start']
                    params['sort'] = 'desc'
            elif 'start' not in params and 'end' not in params:
                raise ValueError("Slice formation must be based on datetime/timedelta values.")



            # handle row limit if included
            if isinstance(s.step, int):
                if s.step < 0:
                    raise ValueError("limit must be non-negative integer or None.")
                params['limit'] = s.step
            elif s.step is not None:
                raise ValueError("limit must be non-negative integer or None.")

        # time conversions
        if 'start' in params:
            params['start'] = iso8601(params['start'])
        if 'end' in params:
            params['end'] = iso8601(params['end'])

        # get data
        resp = self._get(params=params)
        if resp.status_code == 200:
            dat = []
            for e in resp.json()['events']:
                x = {}
                for key in e['event']:
                    x[key] = e['event'][key]
                dat.append(Event(ts=dt64(e['ts']), data=x))
            return dat
        else:
            raise Exception(resp.status_code)

